#ifndef LUM_OBJECT_H
#define LUM_OBJECT_H

/*
  This source is part of the Illumination library
  Copyright (C) 2004  Tim Teulings

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include <list>
#include <map>

#include <Lum/Base/Object.h>
#include <Lum/Base/Size.h>

#include <Lum/Model/Adjustment.h>

#include <Lum/OS/DrawInfo.h>
#include <Lum/OS/Event.h>
#include <Lum/OS/Fill.h>
#include <Lum/OS/Frame.h>
#include <Lum/OS/Window.h>

namespace Lum {

  class Object;

  typedef unsigned long Action;

  /**
    Visitor allows you to enumerate a list of objects.
    You call a special visit-method, handing over an instance of
    Visitor (or a derived class containing additional information).
    The visit method will be called a every object visited.

    This makes it easy to iterate over a collection of objects
    without knowing the internal structure of the collection.
    It follows the iterator design pattern.
   */
  class LUMAPI Visitor
  {
  public:
    virtual ~Visitor();
    virtual bool Visit(Lum::Object* object);
  };

  /**
    Controller, mapping external interaction like keyboard and
    mouse events onto actions supported by the corresponding
    object.

    Normally all objects of a certain type share the same derived
    controller instance.
  */
  class Controller : public Base::Referencable
  {
  public:
    static const Action actionNone = 0xffff; //! Constant value for the "NULL"-action.

  private:
    std::map<std::wstring,Action> keys;

  public:
    virtual ~Controller();

    void RegisterKeyAction(const std::wstring& key, Action action);
    Action GetActionForKey(const std::wstring& key);
    virtual bool DoAction(Lum::Object* object, Action action);
  };

#if defined(LUM_INSTANTIATE_TEMPLATES)
    LUM_EXPTEMPL template class LUMAPI Lum::Base::Reference<Controller>;
#endif

  typedef Lum::Base::Reference<Controller> ControllerRef;

  /**
   \brief The base class for all graphical objects.

   This class is the base class for all graphical objects. It support attributes and methods
   supporting
   - object position and minimal, maximum and current bounds
   - Resizing
   - Drawing and refreshing
   - Keyboard Focus handling
  */
  class LUMAPI Object : public Base::DnDObject
  {
  public:

    /**
     Special Preferences object for graphical objects.
    */
    class LUMAPI Prefs : public Lum::Prefs::Prefs
    {
    public:
      OS::FillRef  background;
      OS::FrameRef frame;
      Base::Size   hSize; //! Value to fixate the horizontal size of the object
      Base::Size   vSize; //! Value to fixate the vertical size of the object

    public:
      Prefs(const std::wstring& name=L"");

      void Initialize();
    };

  protected:
    Object              *parent;        //! The parent object, may be NULL for top level objects.

    Prefs               *prefs;         //! Preferences of this object

    /**
      @name Internal position and dimension.
      The objects internal (without outer frame) minimal, current and maximal dimension.
    */
    //@{
    int                 x;
    int                 y;
    size_t              width;
    size_t              height;
    size_t              minWidth;
    size_t              minHeight;
    size_t              maxWidth;
    size_t              maxHeight;
    //@}

    /**
      @name External position and dimension.
      The objects outer (including outer frame) minimal, current and maximal dimension.
    */
    //@{
    int                 oX;
    int                 oY;
    size_t              oWidth;
    size_t              oHeight;
    size_t              oMinWidth;
    size_t              oMinHeight;
    size_t              oMaxWidth;
    size_t              oMaxHeight;
    //@}

    /**
      @name Initial dimensions.
      The initial dimension of the object.
    */
    //@{
    Base::Size          setWidth,setHeight,
                        setMinWidth,setMinHeight,
                        setMaxWidth,setMaxHeight;
    //@}

    /**
      @name Look.
      Common variables that define the look of the widget.
    */
    //@{
    OS::FillRef         background;    //! The object background. \see Object::SetBackground
    OS::FrameRef        frame;         //! The object frame. \see Object::SetFrame
    //@}

    bool                horizontalFlex,verticalFlex;
    bool                visible;       //! The object is currently visible
    bool                firstDraw;     //! If true, this is the first drawing of the widget
    bool                layouted;      //! The object has already calculated its layout

  private:
    /**
      @name Keyboard focus.
      A number of flags that handle keyboard focus behaviour.
    */
    //@{
    bool                hasFocus;      //! The object currently has the focus
    bool                canFocus;      //! The object can handle keyboard focus
    bool                showFocus;     //! The object does not have the focus, but should draw it anyway
    bool                requestsFocus; //! The object request the keyboard focus
    bool                requestsLabelFocus; //! The control requests that the label should draw focus, too.
    bool                requestsKeyboard; //! The control requests keyboard input
    //@}

    bool                canDrawMouseActive; //! Does draw itself different if is active
    bool                mouseActive;   //! Is currently activated by mouse

  public:
    bool                inited;        //! Has been initialized
    bool                shortcutAlways;

    mutable OS::Window* window;        //! The OS::Window this object belongs to.
    OS::Window*         menu;          //! The window for a popup menu. \see Object::SetMenu
    OS::Window*         help;

    Object              *label;        //! The assigned label. May be NULL. \see Object::SetLabel.
    Object              *labeling;     //! The object we are label of. May be NULL. \see Object::SetLabel.

    ControllerRef       controller;    //! The assigned Controller. May be NULL. \see Object::SetController.

  private:
    Object(const Object& object);

  protected:
    Object();

    /**
      @name Internal position and size
      Methods returning the internal position and size. The internal dimension is defined by the
      size without the outer frame around the object.
    */
    //@{
    virtual int GetX() const;
    virtual int GetY() const;
    virtual size_t GetWidth() const;
    virtual size_t GetHeight() const;
    //@}

  public:
    virtual ~Object();

    virtual void SetPrefs(Prefs* prefs);

    virtual void SetController(Controller* controller);

    /**
      @name position and size
      Methods returning the (outer) position and size, this is the dimension including the
      the frame surounding the object.
    */
    //@{
    virtual int GetOX() const;
    virtual int GetOY() const;
    virtual size_t GetOWidth() const;
    virtual size_t GetOHeight() const;
    virtual size_t GetOMinWidth() const;
    virtual size_t GetOMaxWidth() const;
    virtual size_t GetOMinHeight() const;
    virtual size_t GetOMaxHeight() const;
    //@}

    /**
      @name Baseline support
      Method that return information about the vertical position
      of the textual baseline of the object.
    */
    //@{
    virtual bool HasBaseline() const;
    virtual size_t GetBaseline() const;
    //@}

    /**
      @name Size setter
      Methods setting the minimal, current and maximum sizeof the object.
    */
    //@{
    virtual void SetWidth(Base::Size::Mode mode, size_t value);
    virtual void SetHeight(Base::Size::Mode mode, size_t value);
    virtual void SetMinWidth(Base::Size::Mode mode, size_t value);
    virtual void SetMinHeight(Base::Size::Mode mode, size_t value);
    virtual void SetMaxWidth(Base::Size::Mode mode, size_t value);
    virtual void SetMaxHeight(Base::Size::Mode mode, size_t value);
    //@}

    /**
      @name Methods for manipulation object resize behaviour
    */
    //@{
    virtual void SetFlex(bool horizontal, bool vertical);
    virtual bool HorizontalFlex() const;
    virtual bool VerticalFlex() const;

    virtual bool CanGrowHorizontal() const;
    virtual bool CanGrowVertical() const;
    virtual bool CanShrinkHorizontal() const;
    virtual bool CanShrinkVertical() const;
    virtual bool CanResize(bool grow, bool horiz) const;
    //@}

    /**
      @name Methods for accessing OS structures
    */
    //@{
    virtual void SetWindow(OS::Window* window);
    virtual OS::Window* GetWindow() const;

    virtual OS::DrawInfo* GetDrawInfo() const;
    //@}

    /**
      @name Methods for accessing the object hierachie
    */
    //@{
    virtual void SetParent(Object* parent);
    virtual Object* GetParent() const;
    //@}

    /**
      @name Methods for accessing assigned popup windows
    */
    //@{
    virtual void SetHelp(OS::Window* window);
    virtual OS::Window* GetHelp() const;

    virtual void SetMenu(OS::Window* window);
    virtual OS::Window* GetMenu() const;
    //@}

    /**
      @name Focus manipulation methods
    */
    //@{
    virtual bool HasFocus() const;
    virtual bool CanFocus() const;
    virtual bool DisplayFocus() const;
    virtual bool HandlesFocus() const;
    virtual void SetRequestLabelFocus(bool request);
    virtual bool RequestsLabelFocus() const;

    virtual void SetCanFocus(bool can);
    virtual void SetShowFocus(bool show);

    virtual void RequestFocus();
    virtual void UnrequestFocus();
    virtual bool RequestedFocus() const;

    virtual void SetRequestsKeyboard(bool request);
    virtual bool RequestsKeyboard() const;

    virtual Object* GetFocusFirst();
    virtual Object* GetFocusLast();
    virtual Object* GetFocusNext(Object* object);
    virtual Object* GetFocusPrevious(Object* object);

    virtual void DrawFocus();
    virtual void HideFocus();
    virtual void LeaveFocus();

    virtual void CatchedFocus();
    virtual void RecatchedFocus();
    virtual void LostFocus();

    virtual void SetLabelObject(Object* label);
    //@}

    virtual bool VisitChildren(Visitor &visitor, bool onlyVisible);


    virtual void CalcSize();

    virtual bool IsInited() const;
    virtual bool IsVisible() const;
    virtual bool ShortcutAlways() const;

    /**
      @name Object resizing and placement methods
    */
    //@{
    virtual void Move(int x, int y);
    virtual void MoveResize(int x, int y, size_t width, size_t height);
    virtual void Resize(size_t width, size_t height);
    virtual void ResizeWidth(size_t width);
    virtual void ResizeHeight(size_t height);
    //@}

    virtual void SetRelayout();
    virtual bool IsLayouted() const;
    virtual void Layout();

    /**
      @name Object point inclusion methods
    */
    //@{
    virtual bool OIntersect(int &x, int &y, size_t &width, size_t &height) const;
    virtual bool Intersect(int &x, int &y, size_t &width, size_t &height) const;
    virtual bool PointIsIn(int x, int y) const;
    virtual bool PointIsIn(const OS::MouseEvent& event) const;
    virtual bool MouseIsIn() const;
    //@}

    /**
      @name Control background manipulation and drawing methods
    */
    //@{
    virtual void SetBackground(OS::Fill* background);

    virtual void DrawBackground(int x, int y, size_t w, size_t h);
    virtual void DrawParentBackground(int x, int y, size_t w, size_t h);
    virtual void DrawBackgroundWithDrawInfo(OS::DrawInfo* draw,
                                            int x, int y, size_t w, size_t h);
    virtual void DrawParentBackgroundWithDrawInfo(OS::DrawInfo* draw,
                                                  int x, int y, size_t w, size_t h);
    virtual void DrawBackgroundRelative(OS::DrawInfo* draw,
                                        OS::Fill* background,
                                        int x, int y, size_t w, size_t h);
    virtual void SetFrame(OS::Frame* frame);
    virtual OS::Frame* GetFrame() const;

    virtual OS::Color GetTextColor(OS::DrawInfo *draw) const;
    //@}

    /**
      @name Object drawing methods
    */
    //@{
    virtual void Draw(int x, int y, size_t w, size_t h);
    virtual void Redraw();
    virtual void Redraw(int x, int y, size_t w, size_t h);
    virtual void Hide();
    //@}

    /**
      @name MouseActive methods
    */
    //@{
    virtual void SetCanDrawMouseActive(bool canDraw);
    virtual bool CanDrawMouseActive() const;
    virtual void MouseActive();
    virtual void MouseInactive();
    virtual bool IsMouseActive() const;
    //@}

    /**
      @name Mouse and keyboard input callbacks
    */
    //@{
    virtual bool HandleMouseEvent(const OS::MouseEvent& event);
    virtual bool HandleKeyEvent(const OS::KeyEvent& event);
    //@}
  };

  /**
    New baseclasse that distinguishes between objects which
    control data and those which are used to control
    layout.
  */
  class LUMAPI Control : public Object
  {
  private:
    Base::ModelRef model;

  public:
    ~Control();

    virtual bool HandlesFocus() const;

    // Setter
    virtual bool SetModel(Base::Model* model);

    // Getter
    Base::Model* GetModel() const;

    void Resync(Base::Model* model, const Base::ResyncMsg& msg);
  };

  /**
    Objects that can scoll their content
  */
  class LUMAPI Scrollable : public Control
  {
  protected:
    Model::AdjustmentRef hAdjustment;
    Model::AdjustmentRef vAdjustment;

  public:
    Scrollable();
    ~Scrollable();

    Model::Adjustment* GetHAdjustment() const;
    Model::Adjustment* GetVAdjustment() const;

    virtual void CalculateDimension();
  };

  /**
    Objects that have child controls.
  */
  class LUMAPI Group : public Object
  {
  protected:
    virtual void PreDrawChild(Object* child);
    virtual void PostDrawChild(Object* child);

  public:
    void Draw(int x, int y, size_t w, size_t h);
    void Hide();
  };

  /**
    A implementation of Group using a simple list for holding all its children.
  */
  class LUMAPI List : public Group
  {
  protected:
    std::list<Object*> list; /** A linked list of all children */

  public:
    ~List();

    bool VisitChildren(Visitor &visitor, bool onlyVisible);

    virtual void Add(Object* object);
  };

  /**
    A Component is an object that behaves as a normal object but just
    wraps a Group as a container for a complex object structure.
  */
  class LUMAPI Component : public Object
  {
  protected:
    Group *container;

  public:
    Component();
    ~Component();

    bool VisitChildren(Visitor &visitor, bool onlyVisible);

    void Layout();
    void Draw(int x, int y, size_t w, size_t h);
    void Hide();
  };

  /**
    A ControlComponent is an object that behaves as a normal control but just
    wraps a Group as a container for a complex object structure.
  */
  class LUMAPI ControlComponent : public Control
  {
  protected:
    Group *container;

  public:
    ControlComponent();
    ~ControlComponent();

    bool VisitChildren(Visitor &visitor, bool onlyVisible);

    void Layout();
    void Draw(int x, int y, size_t w, size_t h);
    void Hide();
  };
}

#endif
