/*
  This source is part of the Illumination library
  Copyright (C) 2004  Tim Teulings

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include <Lum/Header.h>

#include <algorithm>

#include <Lum/Base/Size.h>

#include <Lum/Text.h>

namespace Lum {

  static Header::Prefs *prefs=new Header::Prefs();

  Header::Prefs::Prefs()
  : Object::Prefs(L"Header")
  {
    // no code
  }

  void Header::Prefs::Initialize()
  {
    Object::Prefs::Initialize();

    background=OS::display->GetFill(OS::Display::columnBackgroundFillIndex);
    columnFrame=OS::display->GetFrame(OS::Display::columnFrameIndex);
    columnLeftFrame=OS::display->GetFrame(OS::Display::columnLeftFrameIndex);
    columnMiddleFrame=OS::display->GetFrame(OS::Display::columnMiddleFrameIndex);
    columnRightFrame=OS::display->GetFrame(OS::Display::columnRightFrameIndex);
  }

  class HeaderEntry : public Object
  {
  private:
    Object *label;

  public:
    HeaderEntry(const std::wstring& label, size_t current, size_t max)
    : label(new Lum::Text(label))
    {
      SetPrefs(Lum::prefs);

      SetCanDrawMouseActive(true);

      if (max==1) {
        SetFrame(dynamic_cast<Header::Prefs*>(prefs)->columnFrame);
      }
      else if (current==1) {
        SetFrame(dynamic_cast<Header::Prefs*>(prefs)->columnLeftFrame);
      }
      else if (current==max) {
        SetFrame(dynamic_cast<Header::Prefs*>(prefs)->columnRightFrame);
      }
      else {
        SetFrame(dynamic_cast<Header::Prefs*>(prefs)->columnMiddleFrame);
      }

      SetBackground(dynamic_cast<Header::Prefs*>(prefs)->background);

      this->label->SetFlex(true,true);
      this->label->SetParent(this);
    }

    ~HeaderEntry()
    {
      delete label;
    }

    void SetLabel(Object *label)
    {
      delete label;
      this->label=label;

      Redraw();
    }

    void CalcSize()
    {
      label->CalcSize();

      minWidth=1;
      minHeight=label->GetOHeight()+frame->minHeight;

      width=label->GetOWidth()+frame->minWidth+
            2*OS::display->GetSpaceHorizontal(OS::Display::spaceObjectBorder);
      height=minHeight;

      Object::CalcSize();
    }

    void Layout()
    {
      label->MoveResize(x+OS::display->GetSpaceHorizontal(OS::Display::spaceObjectBorder),
                        y,
                        width-2*OS::display->GetSpaceHorizontal(OS::Display::spaceObjectBorder),
                        height);

      Object::Layout();
    }

    void Draw(int x, int y, size_t w, size_t h)
    {
      OS::DrawInfo *draw=GetDrawInfo();

      draw->activated=IsMouseActive();

      Object::Draw(x,y,w,h);

      if (!Intersect(x,y,w,h)) {
        draw->activated=false;
        return;
      }

      /* --- */

      draw->PushClipBegin(x,y,w,h);
      draw->SubClipRegion(label->GetOX(),label->GetOY(),
                          label->GetOWidth(),label->GetOHeight());
      draw->PushClipEnd();
      DrawBackground(x,y,w,h);
      draw->PopClip();

      label->Draw(x,y,w,h);

      draw->activated=false;
    }
  };

  Header::Header()
  : model(NULL),adjustment(NULL)
  {
    SetPrefs(Lum::prefs);
  }

  Header::~Header()
  {
    for (size_t x=0; x<objects.size(); x++) {
      delete objects[x];
    }

    if (adjustment.Valid()) {
      UnattachModel(adjustment->GetTopModel());
    }

    if (model.Valid()) {
      UnattachModel(model);
    }
  }

  bool Header::VisitChildren(Visitor &visitor, bool /*onlyVisible*/)
  {
    for (size_t x=0; x<objects.size(); x++) {
      if (!visitor.Visit(objects[x])) {
        return false;
      }
    }

    return true;
  }

  void Header::SetAdjustment(Model::Adjustment* adjustment)
  {
    if (this->adjustment.Valid()) {
      UnattachModel(this->adjustment->GetTopModel());
    }

    this->adjustment=adjustment;

    AttachModel(this->adjustment->GetTopModel());
  }

  void Header::CalcSize()
  {

    if (!model.Valid()) {
      width=1;
      height=OS::display->GetFont()->height+
             dynamic_cast<Prefs*>(prefs)->columnFrame->minHeight/*+
             2*OS::display->GetSpaceVertical(OS::Display::spaceObjectBorder)*/;
    }
    else {
      objects.resize(model->GetColumns());

      for (size_t x=0; x<model->GetColumns(); x++) {
        objects[x]=new HeaderEntry(model->GetLabel(x),x+1,model->GetColumns());
        objects[x]->SetFlex(true,true);
        objects[x]->SetParent(this);
        objects[x]->CalcSize();

        if (model->GetColumnWidth(x)==0) {
          model->SetColumnWidth(x,objects[x]->GetOWidth());
        }
        height=std::max(height,objects[x]->GetOMinHeight());
      }

      width=0;
    }

    Control::CalcSize();
  }

  void Header::DrawHeader(OS::DrawInfo* draw, size_t startColumn)
  {
    if (model.Valid()) {
      int    xPos;
      size_t columnsWidth=0;

      if (adjustment.Valid()) {
        xPos=x-adjustment->GetTop()+1;
      }
      else {
        xPos=x;
      }

      draw->PushClip(x,y,width,height);

      for (size_t count=0; count<objects.size(); count++) {
        if (count>=startColumn &&
            !(xPos>=x+(int)width) &&
            !(xPos+(int)model->GetColumnWidth(count)-1<x)) {

          if (model->IsSortColumn(count)) {
            draw->selected=true;
          }

          objects[count]->MoveResize(xPos,y,model->GetColumnWidth(count),height);
          objects[count]->Draw(x,y,width,height);

          draw->selected=false;
        }

        xPos+=model->GetColumnWidth(count);
        columnsWidth+=model->GetColumnWidth(count);
      }

      if (xPos<x+(int)width) {
        OS::display->GetFill(OS::Display::columnEndBackgroundFillIndex)->Draw(draw,
                                                                              xPos,y,x+width-xPos+1,height,
                                                                              xPos,y,x+width-xPos+1,height);
        OS::display->GetFrame(OS::Display::columnEndFrameIndex)->Draw(draw,
                                                                      xPos,y,x+width-xPos+1,height);
      }

      draw->PopClip();

      if (adjustment.Valid()) {
        adjustment->SetDimension(width,columnsWidth);
      }
    }
    else {
      DrawBackground(x,y,width,height);
    }
  }

  bool Header::SetModel(Base::Model* model)
  {
    this->model=dynamic_cast<Model::Header*>(model);

    Control::SetModel(this->model);

    return this->model.Valid();
  }

  void Header::Draw(int x, int y, size_t w, size_t h)
  {
    Control::Draw(x,y,w,h);

    if (!Intersect(x,y,w,h)) {
      return;
    }

    /* --- */

    DrawHeader(GetDrawInfo(),0);
  }

  bool Header::HandleMouseEvent(const OS::MouseEvent& event)
  {
    if (!visible || !model.Valid() || !model->IsEnabled()) {
      return false;
    }

    if (event.type==OS::MouseEvent::down && PointIsIn(event) && event.button==OS::MouseEvent::button1) {
      int pos;

      pos=x;
      if (adjustment.Valid()) {
        pos-=adjustment->GetTop()-1;
      }

      for (size_t current=0; current<objects.size(); current++) {
        pos+=model->GetColumnWidth(current);

        if (abs(pos-event.x)<=2) {
          if (GetWindow()->IsDoubleClicked()) {
            model->RequestFit(current);
          }
          else {
            offset=event.x-pos;
            grabStart=pos-model->GetColumnWidth(current);
            grabColumn=current;
          }

          return true;
        }
      }
    }
    else if (event.type==OS::MouseEvent::move && event.IsGrabed()) {
      if (event.x-offset>grabStart+(int)OS::display->GetSpaceHorizontal(OS::Display::spaceObjectSensibility)) {
        model->SetColumnWidth(grabColumn,event.x-offset-grabStart);
      }
    }

    return false;
  }

  void Header::Resync(Base::Model* model, const Base::ResyncMsg& msg)
  {
    const Model::Header::ResizedColumn *resized;

    if (!visible) {
      return;
    }

    if (model==this->model) {
      if (IsVisible() && (resized=dynamic_cast<const Model::Header::ResizedColumn*>(&msg))!=NULL) {
        DrawHeader(GetDrawInfo(),resized->column);
      }
      else {
        Redraw();
      }
    }
    else if (adjustment.Valid() && model==adjustment->GetTopModel()) {
      Redraw();
    }
  }
}
